@@ -5,6 +5,7 @@ from pathlib import Path
 import pydicom
 import shutil
 import time
+import timeit
 
 import skrt.core
 from skrt.image import Image
@@ -50,6 +51,18 @@ class Study(skrt.core.Archive):
             load=False
         )
 
+        if hasattr(self, 'plan_types'):
+            for plans in self.plan_types.values():
+                for plan in plans:
+                    self.link_plan_to_doses(plan)
+
+        if 'ct' in self.image_types and 'cthd' in self.image_types:
+            if (len(self.image_types['ct']) == 1
+                and len(self.image_types['cthd']) == 1):
+                self.cthd_images[0].structure_sets = []
+                for ss in self.ct_images[0].structure_sets:
+                    self.cthd_images[0].add_structure_set(ss)
+                    ss.set_image(self.cthd_images[0])
         #  self.load_dose_or_plan()
 
     def load_images(self):
@@ -65,17 +78,17 @@ class Study(skrt.core.Archive):
             subdir = os.path.basename(file.path)
             if subdir in special_dirs:
                 continue
-            self.image_types[subdir] = {}
+            self.image_types[subdir.lower()] = {}
 
             # Get Image objects in this subdir
             attr = f"{subdir.lower()}_images"
             images = self.create_objects(dtype=Image, subdir=subdir, load=False)
             for im in images:
-                im.image_type = subdir
+                im.image_type = subdir.lower()
 
             # Store the images
             setattr(self, attr, images)
-            self.image_types[subdir] = images
+            self.image_types[subdir.lower()] = images
 
     def load_from_subdir(self, dtype, subdir, attr_name, **kwargs):
         """Create objects of type <dtype> from each directory in <subdir> and 
@@ -102,13 +115,18 @@ class Study(skrt.core.Archive):
         if not os.path.exists(obj_dir):
             return
 
+        # Initialise dictionary of all objects created.
+        obj_types_name = f'{attr_name[:-1]}_types'
+        setattr(self, obj_types_name, {})
+        obj_types = getattr(self, obj_types_name)
+
         # Search subdirectories (each corresponds to an image type)
         for im_type_dir in os.listdir(obj_dir):
 
-            im_type = im_type_dir
+            im_type = im_type_dir.lower()
             if subdir in ['RTDOSE', 'RTPLAN'] \
                     and im_type not in self.image_types:
-                im_type = 'CT'
+                im_type = 'ct'
             subpath = f"{subdir}/{im_type_dir}"
 
             all_objs = []
@@ -128,7 +146,7 @@ class Study(skrt.core.Archive):
                 # Look for an image matching the timestamp of this archive
                 if im_type in self.image_types:
                     image_types = self.image_types[im_type]
-                    if im_type != im_type_dir:
+                    if im_type != im_type_dir.lower():
                         obj_to_match = objs[0]
                     else:
                         obj_to_match = archive
@@ -137,6 +155,7 @@ class Study(skrt.core.Archive):
                         for obj in objs:
                             if hasattr(obj, "set_image"):
                                 obj.set_image(image)
+
                     if ss is not None:
                         for obj in objs:
                             if hasattr(obj, "set_structure_set"):
@@ -147,9 +166,11 @@ class Study(skrt.core.Archive):
 
             # Create attribute for objects of this type
             if len(all_objs):
-                setattr(self, f"{im_type_dir.lower()}_{attr_name}", all_objs)
+                key = f"{im_type_dir.lower().replace('-', '_')}"
+                setattr(self, f"{key}_{attr_name}", all_objs)
+                obj_types[key] = all_objs
 
-    def add_image(self, im, image_type="CT"):
+    def add_image(self, im, image_type="ct"):
         '''Add a new image of a given image type.'''
 
         # Ensure we have a list of this type of image
@@ -167,16 +188,16 @@ class Study(skrt.core.Archive):
 
         # Also add to dict of images of each type
         if image_type not in self.image_types:
-            self.image_types[image_type] = [im]
+            self.image_types[image_type.lower()] = [im]
         else:
-            self.image_types[image_type].append(im)
+            self.image_types[image_type.lower()].append(im)
 
         # Ensure image has a timestamp
         if not im.timestamp:
             im.timestamp = skrt.core.generate_timestamp()
 
         # Ensure corresponding structure list exists
-        struct_subdir = f"RTSTRUCT/{image_type}"
+        struct_subdir = f"RTSTRUCT/{image_type.upper()}"
         struct_name = f"{image_type.lower()}_structure_sets"
         if not hasattr(self, struct_name):
             setattr(self, struct_name, [])
@@ -439,6 +460,300 @@ class Study(skrt.core.Archive):
 
         return plan_dose
 
+    def link_plan_to_doses(self, plan):
+        '''
+        Link plan to doses derived from it.
+
+        **Parameter:**
+
+        plan : skrt.dose.Plan
+            Plan object for which dose associations are to be determined.
+        '''
+
+        plan_uid = pydicom.dcmread(plan.path, force=True).SOPInstanceUID
+        doses = []
+        if hasattr(self, 'dose_types'):
+            for value in self.dose_types.values():
+                doses.extend(value)
+        for dose in doses:
+            dose_ds = pydicom.dcmread(dose.path, force=True)
+            if hasattr(dose_ds, 'ReferencedRTPlanSequence'):
+                for referenced_plan in dose_ds.ReferencedRTPlanSequence:
+                    if plan_uid == referenced_plan.ReferencedSOPInstanceUID:
+                        dose.set_plan(plan)
+
+    def save_images_as_nifti(self, out_dir='.', image_types=None, times=None,
+            verbose=True, image_size=None, voxel_size=None, fill_value=-1024,
+            bands=None, require_structure_set=None):
+        '''
+        Save study's image data as nifti files.
+
+        **Parameters:**
+
+        out_dir - str, default='.'
+            Directory to which nifti files will be saved.
+
+        image_types - list/str/None, default=None
+            Images types to be saved: None to save all, otherwise a list
+            of image types to save, or a string specifying a single image
+            type to save.
+
+        times : dict, default=None
+            Dictionary where the keys are image types and the values are
+            lists of timestamp indices for the images to be saved,
+            0 being the earliest and -1 being the most recent.  If set to
+            None, all images are saved.
+
+        verbose - bool, default=True
+           Flag indicating whether to report progress.
+
+        image_size - tuple, default=None
+            Image dimensions (dx, dy, dz) in voxels for image resizing
+            prior to writing.  If None, original image size is kept.
+
+        voxel_size - tuple, default=None
+            Voxel dimensions (dx, dy, dz) in mm for image resizing
+            prior to writing.  If None, original voxel size is kept.
+
+        fill_value - int/float, default=-1024
+            Value used when extrapolating image outside data area>
+
+        bands - dict, default=None
+            Nested dictionary of value bandings to be applied before
+            image saving.  The primary key defines the type of image to
+            which the banding applies.  Secondary keys specify band limits,
+            and associated values indicte the values to be assigned.
+            For example:
+
+            - bands{'mvct' : {-100: -1024, 100: 0, 1e10: 1024}}
+
+            will band an image of type 'mvct':
+
+            - value <= -100 => -1024;
+            - -100 < value <= 100 => 0;
+            - 100 < value <= 1e10 => 1024.
+
+        require_structure_set - list, default=None
+            List of image types for which data are to be written only
+            if there is an associated structure set.
+        '''
+
+        # Set defaults.
+        if bands is None:
+            bands = {}
+        if require_structure_set is None:
+            require_structure_set = []
+
+        # Obtain full path to output directory.
+        out_dir = skrt.core.fullpath(out_dir)
+
+        # Obtain set of image types to be saved.
+        save_types = set(self.image_types)
+        if image_types is not None:
+            if isinstance(image_types, str):
+                image_types = {image_types}
+            save_types = save_types.intersection(set(image_types))
+        if verbose:
+            if save_types:
+                print(f"Saving imaging data to '{out_dir}':")
+            else:
+                print(f"No data to save for output directory '{out_dir}'")
+
+        # Loop over image types to be saved.
+        for save_type in save_types:
+            images = sorted(self.image_types[save_type])
+
+            # Only consider images with requested indices.
+            if times:
+                if save_type in times and times[save_type]:
+                    images = [images[idx] for idx in times[save_type]]
+
+            # Loop over images.
+            for idx, image in enumerate(images):
+
+                # Skip image if associated structure set required but missing.
+                if (save_type in require_structure_set and
+                        not image.structure_sets):
+                    continue
+
+                # Identify imaging machine.
+                suffix = (image.get_machine() if image.get_machine()
+                        else 'Unknown')
+
+                # Clone image object for writing.
+                # If data manipulation is required, this will be
+                # applied to the clone rather than to the original.
+                out_image = Image(image)
+
+                # Resize to required image size and voxel size.
+                if image_size is not None or voxel_size is not None:
+                    out_image.resize(image_size=image_size,
+                            voxel_size=voxel_size, fill_value=fill_value)
+
+                # Apply banding.
+                if save_type in bands:
+                    if bands[save_type]:
+                        out_image.apply_bandings(bands[save_type])
+
+                # Define output path and write image.
+                outname = (f'{out_dir}/{save_type.upper()}'
+                           f'_{image.timestamp}_{idx+1:03}'
+                           f'_{suffix}.nii.gz')
+                out_image.write(outname=outname, verbose=verbose)
+
+    def save_structure_sets_as_nifti(self, out_dir='.',
+            image_types=None, times=None, files=None,
+            verbose=True, image_size=None, voxel_size=None, roi_names=None,
+            force_roi_nifti=False, bilateral_names=None):
+
+        '''
+        Save study's structure-set data as nifti files.
+
+        **Parameters:**
+
+        out_dir - str, default='.'
+            Directory to which nifti files will be saved.
+
+        image_types - list/str/None, default=None
+            Images types for which structure sets are to be saved: None to
+            save for all, otherwise a list of image types, or a string
+            specifying a single image type.
+
+        times : dict, default=None
+            Dictionary where the keys are image types and the values are
+            lists of timestamp indices for the images for which structure
+            sets are to be saved, 0 being the earliest and -1 being the
+            most recent.  If set to None, structure sets are saved for
+            all images.
+
+        files : dict, default=None
+            Dictionary where the keys are image types and the values are
+            lists of file indices for structure sets to be saved for
+            a given image, 0 being the earliest and -1 being the most recent.
+            If set to None, all of an image's structure sets are saved.
+
+        verbose - bool, default=True
+           Flag indicating whether to report progress.
+
+        image_size - tuple, default=None
+            Image dimensions (dx, dy, dz) in voxels for image resizing
+            prior to writing.  If None, original image size is kept.
+
+        voxel_size - tuple, default=None
+            Voxel dimensions (dx, dy, dz) in mm for image resizing
+            prior to writing.  If None, original voxel size is kept.
+
+        roi_names : dict, default=None
+            Dictionary of names for renaming ROIs, where the keys are new 
+            names and values are lists of possible names of ROIs that should
+            be assigned the new name. These names can also contain wildcards
+            with the '*' symbol.
+
+        force_roi_nifti : bool, default=False
+            If True, force writing of dummy NIfTI file for all named ROIs
+            not found in structure set.  If False, named ROIs not found in
+            structure set are disregarded.
+
+        bilateral_names : list, default=None
+            List of names of ROIs that should be split into
+            left and right parts.
+        '''
+
+        # Obtain full path to output directory.
+        out_dir = skrt.core.fullpath(out_dir)
+
+        # Initialise list or names for bilateral ROIs
+        if bilateral_names is None:
+            bilateral_names = []
+
+        # Obtain set of image types for which structure sets are to be saved.
+        save_types = set(self.image_types)
+        if image_types is not None:
+            if isinstance(image_types, str):
+                image_types = {image_types}
+            save_types = save_types.intersection(set(image_types))
+        if verbose:
+            if save_types:
+                print(f"Saving structure_set data to '{out_dir}':")
+            else:
+                print(f"No data to save for output directory '{out_dir}'")
+
+        # Loop over image types for which structure sets are to be saved.
+        for save_type in save_types:
+            if save_type not in self.image_types:
+                continue
+            images = sorted(self.image_types[save_type])
+
+            # Only consider images with requested indices.
+            if times:
+                if save_type in times and times[save_type]:
+                    images = [images[idx] for idx in times[save_type]]
+
+            # Loop over images.
+            for image in images:
+                structure_sets = sorted(image.structure_sets)
+
+                # Only consider structure sets with requested indices.
+                if files and structure_sets:
+                    if save_type in files and files[save_type]:
+                        structure_sets = [structure_sets[idx]
+                                for idx in files[save_type]]
+
+                # Loop over structure sets
+                idx = 0
+                for structure_set in structure_sets:
+                    # Filter ROIs
+                    ss = structure_set.filtered_copy(names=roi_names,
+                            keep_renamed_only=True)
+
+                    # Resize associated image if needed.
+                    if ss.get_roi_names() or force_roi_nifti:
+                        if image_size is not None or voxel_size is not None:
+                            im = Image(ss.image)
+                            im.resize(image_size=image_size,
+                                    voxel_size=voxel_size)
+                            ss.set_image(im)
+
+                    # Create list of ROIs to be saved.
+                    if roi_names:
+                        save_names = list(roi_names)
+                    else:
+                        save_names = list(ss.get_roi_names())
+
+                    # Split ROI into left and right parts if needed.
+                    for roi_name in bilateral_names:
+                        save_names.remove(roi_name)
+                        save_names.append(f'{roi_name}_left')
+                        save_names.append(f'{roi_name}_right')
+                        if roi_name in ss.get_roi_names():
+                            ss_tmp = ss.get_roi(roi_name).split_in_two()
+                            for roi in ss_tmp.get_rois():
+                                ss.add(roi)
+
+                    # Loop over ROIs.
+                    idx_add = 0
+                    for roi_name in save_names:
+                        # Define output path
+                        outname = (f'RTSTRUCT_{save_type.upper()}'
+                                   f'_{image.timestamp}_{idx+1:03}'
+                                   f'_{roi_name}.nii.gz')
+                        if roi_name in ss.get_roi_names():
+                            # Save ROI mask.
+                            roi = ss.get_roi(roi_name)
+                            roi.write(outname=outname, outdir=out_dir,
+                                    ext='nii.gz', verbose=verbose)
+                            idx_add = 1
+                        elif force_roi_nifti:
+                            # Save dummy mask for ROI not found
+                            # in structure set.
+                            im = Image(ss.image)
+                            shape = im.get_data().shape
+                            im.data = np.zeros(shape)
+                            outname = f'{out_dir}/{outname}'
+                            ss.image.write(outname=outname, verbose=verbose)
+                            idx_add = 1
+                    idx += idx_add
+
 
 class Patient(skrt.core.PathData):
     """
@@ -448,6 +763,9 @@ class Patient(skrt.core.PathData):
 
     def __init__(self, path="", exclude=["logfiles"]):
 
+        # Record start time
+        tic = timeit.default_timer()
+
         # Set path and patient ID
         if path is None:
             path = os.getcwd()
@@ -466,8 +784,12 @@ class Patient(skrt.core.PathData):
                                 self.create_objects(dtype=Study, subdir=subdir)
                             )
 
+        # Record end time, then store initialisation time.
+        toc = timeit.default_timer()
+        self.init_time = (toc - tic)
+
     def add_study(self, subdir='', timestamp=None, images=None, 
-                  image_type="CT"):
+                  image_type="ct"):
         '''Add a new study.'''
     
         # Create empty Study object
@@ -478,7 +800,7 @@ class Patient(skrt.core.PathData):
         # Add images
         if images:
             for im in images:
-                s.add_image(im, image_type)
+                s.add_image(im, image_type.lower())
 
         # Add to studies list
         self.studies.append(s)
@@ -539,7 +861,8 @@ class Patient(skrt.core.PathData):
         obj = None
         if self.studies:
             obj = getattr(
-                self.studies[0], f"{self.studies[0].image_types[0].lower()}_images"
+                self.studies[0],
+                f"{self.studies[0].image_types[0].lower()}_images"
             )[-1]
 
         # Read demographic info from the object
@@ -756,7 +1079,7 @@ class Patient(skrt.core.PathData):
                 if image_type in to_ignore:
                     continue
 
-                image_type_dir = os.path.join(study_dir, image_type)
+                image_type_dir = os.path.join(study_dir, image_type.upper())
                 if not os.path.exists(image_type_dir):
                     os.mkdir(image_type_dir)
 
@@ -778,7 +1101,7 @@ class Patient(skrt.core.PathData):
                     if os.path.exists(outname) and not overwrite:
                         continue
                     Image.write(im, outname, patient_id=self.id,
-                                modality=image_type, root_uid=root_uid)
+                                modality=image_type.upper(), root_uid=root_uid)
 
                     # Write associated structure sets
                     self.write_non_image_data(im, image_type, im_timestamp,
@@ -819,7 +1142,7 @@ class Patient(skrt.core.PathData):
                 item_dir = os.path.join(outdir, modality, item_subpath)
             else:
                 item_dir = os.path.join(
-                        outdir, modality, image_type, im_timestamp)
+                        outdir, modality, image_type.upper(), im_timestamp)
             filename = f'{modality}_{item.timestamp}'
 
             # Ensure it exists
@@ -857,7 +1180,8 @@ def find_matching_object(obj, possible_matches):
 
     # If no timestamp match, try matching on SOP Instance UID
     if issubclass(type(obj), Dose):
-        ds_obj = obj.get_dicom_dataset()
+        # ds_obj = obj.get_dicom_dataset()
+        ds_obj = pydicom.dcmread(obj.path, force=True)
         if hasattr(ds_obj, 'ReferencedImageSequence'):
             # Omit part of UID after final dot,
             # to be insenstive to slice/frame considered.
@@ -865,7 +1189,8 @@ def find_matching_object(obj, possible_matches):
                     ds_obj.ReferencedImageSequence[-1]
                     .ReferencedSOPInstanceUID.split('.')[:-1])
             for match in possible_matches:
-                ds_match = match.get_dicom_dataset()
+                # ds_match = match.get_dicom_dataset()
+                ds_match = pydicom.dcmread(match.files[0].path, force=True)
                 if hasattr(ds_match, 'SOPInstanceUID'):
                     sop_instance_uid = '.'.join(
                             ds_match.SOPInstanceUID.split('.')[:-1])
@@ -873,13 +1198,15 @@ def find_matching_object(obj, possible_matches):
                         return (match, None)
 
     elif issubclass(type(obj), Plan):
-        ds_obj = obj.get_dicom_dataset()
+        # ds_obj = obj.get_dicom_dataset()
+        ds_obj = pydicom.dcmread(obj.path, force=True)
         if hasattr(ds_obj, 'ReferencedStructureSetSequence'):
             referenced_sop_instance_uid = ds_obj.\
                     ReferencedStructureSetSequence[-1].ReferencedSOPInstanceUID
             for match in possible_matches:
                 for structure_set in match.get_structure_sets():
-                    ds_match = structure_set.get_dicom_dataset()
+                    # ds_match = structure_set.get_dicom_dataset()
+                    ds_match = pydicom.dcmread(structure_set.path, force=True)
                     if hasattr(ds_match, 'SOPInstanceUID'):
                         sop_instance_uid = ds_match.SOPInstanceUID
                         if sop_instance_uid == referenced_sop_instance_uid:
